/*
 * authkeys.c - internal authentication providers
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>

#include "ntp_types.h"
#include "ntp_fp.h"
#include "ntp.h"
#include "ntpd.h"
#include "ntp_string.h"
#include "ntp_malloc.h"
#include "ntp_stdlib.h"
#include "ntp_authprov.h"
#include "ntp_request.h"

/*
 * Structure to store keys in in the hash table.
 */
struct savekey {
	struct savekey *next;
	union {
		long bogon;		/* Make sure nonempty */
		u_char MD5_key[32];	/* MD5 key */
	} k;
	keyid_t keyid;		/* key identifier */
	u_short flags;		/* flags that wave */
	u_long lifetime;	/* remaining lifetime */
	int keylen;		/* key length */
};

#define	KEY_TRUSTED	0x001	/* this key is trusted */
#define	KEY_MD5		0x200	/* this is a MD5 type key */

/*
 * The hash table. This is indexed by the low order bits of the
 * keyid. We make this fairly big for potentially busy servers.
 */
#define	HASHSIZE	64
#define	HASHMASK	((HASHSIZE)-1)
#define	KEYHASH(keyid)	((keyid) & HASHMASK)

#define	MEMINC	12		/* number of new free ones to get */

/*
 * Internal authentication provider, shared between keyfile
 * and autokey backends.
 */
struct internal_authkey_be {
	/* Method dispatch table */
	struct authkey_be be;

	/* Backend parameters */
	char *keyfile;

	/* Key cache */
	keyid_t	cache_keyid;		/* key identifier */
	u_char	*cache_key;		/* key pointer */
	u_int	cache_keylen;		/* key length */
	u_short cache_flags;		/* flags that wave */

	/* Key storage */
	struct savekey *key_hash[HASHSIZE];
	struct savekey *authfreekeys;

	/* Statistics */
	struct info_auth stats;
};

/*
 * internal_authprov_havekey - return one and optionally cache the key, if known and trusted.
 */
static int
internal_authprov_havekey(
	struct authkey_be *be,
	keyid_t keyno,
	int cache
	)
{
	struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;
	struct savekey *sk;

	if (keyno == ibe->cache_keyid) {
		return (1);
	}

	ibe->stats.keyuncached++;
	sk = ibe->key_hash[KEYHASH(keyno)];
	while (sk != 0) {
		if (keyno == sk->keyid)
		    break;
		sk = sk->next;
	}
	if (sk == 0) {
		ibe->stats.keynotfound++;
		return (0);
	} else if (cache == 0) {
		/* Found key, not concerned whether it is trusted or not */
		return (1);
	} else if (!(sk->flags & KEY_TRUSTED)) {
		return (0);
	}
	ibe->cache_keyid = sk->keyid;
	ibe->cache_flags = sk->flags;
	if (sk->flags & KEY_MD5) {
		ibe->cache_key = sk->k.MD5_key;
		ibe->cache_keylen = sk->keylen;
		return (1);
	}
	return (0);
}


/*
 * internal_authprov_moremem - get some more free key structures
 */
static int
internal_authprov_moremem(struct internal_authkey_be *ibe)
{
	struct savekey *sk;
	int i;

	sk = (struct savekey *)calloc(MEMINC, sizeof(struct savekey));
	if (sk == 0)
		return (0);
	
	for (i = MEMINC; i > 0; i--) {
		sk->next = ibe->authfreekeys;
		ibe->authfreekeys = sk++;
	}
	ibe->stats.numfreekeys += MEMINC;
	return (ibe->stats.numfreekeys);
}


/*
 * internal_authprov_trust - declare a key to be trusted/untrusted
 */
static void
internal_authprov_trust(
	struct authkey_be *be,
	keyid_t keyno,
	u_long trust
	)
{
	struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;
	struct savekey *sk;

#ifdef DEBUG
	if (debug > 2)
		printf("authtrust: keyid %08x life %lu\n", keyno, trust);
#endif
	sk = ibe->key_hash[KEYHASH(keyno)];
	while (sk != 0) {
		if (keyno == sk->keyid)
		    break;
		sk = sk->next;
	}

	if (sk == 0 && !trust)
		return;
	
	if (sk != 0) {
		if (ibe->cache_keyid == keyno) {
			ibe->cache_flags = 0;
			ibe->cache_keyid = 0;
		}

		if (trust > 0) {
			sk->flags |= KEY_TRUSTED;
			if (trust > 1)
				sk->lifetime = current_time + trust;
			else
				sk->lifetime = 0;
			return;
		}

		sk->flags &= ~KEY_TRUSTED;
		{
			struct savekey *skp;

			skp = ibe->key_hash[KEYHASH(keyno)];
			if (skp == sk) {
				ibe->key_hash[KEYHASH(keyno)] = sk->next;
			} else {
				while (skp->next != sk)
				    skp = skp->next;
				skp->next = sk->next;
			}
			ibe->stats.numkeys--;

			sk->next = ibe->authfreekeys;
			ibe->authfreekeys = sk;
			ibe->stats.numfreekeys++;
		}
		return;
	}

	if (ibe->stats.numfreekeys == 0) {
	    if (internal_authprov_moremem(ibe) == 0)
		return;
	}

	sk = ibe->authfreekeys;
	ibe->authfreekeys = sk->next;
	ibe->stats.numfreekeys--;

	sk->keyid = keyno;
	sk->flags = KEY_TRUSTED;
	sk->next = ibe->key_hash[KEYHASH(keyno)];
	ibe->key_hash[KEYHASH(keyno)] = sk;
	ibe->stats.numkeys++;
	return;
}


/*
 * authistrusted - determine whether a key is trusted
 */
static int
internal_authprov_istrusted(
       struct authkey_be *be,
	keyid_t keyno
	)
{
        struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;
	struct savekey *sk;

       if (keyno == ibe->cache_keyid)
           return ((ibe->cache_flags & KEY_TRUSTED) != 0);

        ibe->stats.keyuncached++;
        sk = ibe->key_hash[KEYHASH(keyno)];
	while (sk != 0) {
		if (keyno == sk->keyid)
		    break;
		sk = sk->next;
	}
	if (sk == 0) {
                ibe->stats.keynotfound++;
		return (0);
	} else if (!(sk->flags & KEY_TRUSTED)) {
                ibe->stats.keynotfound++;
		return (0);
	}
	return (1);
}

/*
 * internal_authprov_setkey - associate a key with a key ID
*/
static int
internal_authprov_setkey(
        struct authkey_be *be,
	int type,
	keyid_t keyno,
	const u_char *key,
	const int len,
	u_long trust
	)
{
        struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;
	struct savekey *sk;

        if (type != KEY_TYPE_MD5) {
                return (0);
        }
	
	/*
	 * See if we already have the key.  If so just stick in the
	 * new value.
	 */
        sk = ibe->key_hash[KEYHASH(keyno)];
	while (sk != 0) {
		if (keyno == sk->keyid) {
			strncpy((char *)sk->k.MD5_key, (const char *)key,
			    sizeof(sk->k.MD5_key));
			if ((sk->keylen = len) > sizeof(sk->k.MD5_key))
			    sk->keylen = sizeof(sk->k.MD5_key);

			sk->flags |= KEY_MD5;
			if (ibe->cache_keyid == keyno) {
				ibe->cache_flags = 0;
				ibe->cache_keyid = 0;
			}
			return (1);
		}
		sk = sk->next;
	}

	/*
	 * Need to allocate new structure.  Do it.
	 */
	if (ibe->stats.numfreekeys == 0) {
		if (internal_authprov_moremem(ibe) == 0)
		    return (0);
	}

	sk = ibe->authfreekeys;
	ibe->authfreekeys = sk->next;
	ibe->stats.numfreekeys--;

	strncpy((char *)sk->k.MD5_key, (const char *)key,
		sizeof(sk->k.MD5_key));
	if ((sk->keylen = len) > sizeof(sk->k.MD5_key))
	    sk->keylen = sizeof(sk->k.MD5_key);

	sk->keyid = keyno;
	sk->flags = KEY_MD5;
	sk->lifetime = 0;
	if (trust > 0) {
		sk->flags |= KEY_TRUSTED;
		if (trust > 1)
			sk->lifetime = current_time + trust;
	}
	sk->next = ibe->key_hash[KEYHASH(keyno)];
	ibe->key_hash[KEYHASH(keyno)] = sk;
	ibe->stats.numkeys++;
	return (1);
}

/*
 * internal_authprov_getmac - generate message authenticator
 *
 * Returns length of authenticator field, zero if key not found.
 */
static int
internal_authprov_getmac(
	struct authkey_be *be,
	keyid_t keyno,
	u_int32 *pkt,
	int length
	)
{
	struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;

	ibe->stats.encryptions++;

	if (!internal_authprov_havekey(be, keyno, 1)) {
		return (0);
	}

	if (ibe->cache_flags & KEY_MD5) {
		return (MD5authencrypt(ibe->cache_key, ibe->cache_keylen, pkt, length));
	}

	return (0);
}

/*
 * internal_authprov_verifymac - verify message authenticator
 *
 * Returns one if authenticator valid, zero if invalid or key not found.
 */
static int
internal_authprov_verifymac(
	struct authkey_be *be,
	keyid_t keyno,
	u_int32 *pkt,
	int length,
	int size
	)
{
	struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;

	ibe->stats.decryptions++;

	if (!internal_authprov_havekey(be, keyno, 1) || size < 4) {
		return (0);
	}

	if (ibe->cache_flags & KEY_MD5) {
		return (MD5authdecrypt(ibe->cache_key, ibe->cache_keylen, pkt, length, size));
	}

	return (0);
}

/* 
 * internal_authprov_stats - get authentication provider statistics
 */
static int
internal_authprov_stats(
	struct authkey_be *be,
	struct info_auth *ia)
{
	struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;

	memcpy(ia, &ibe->stats, sizeof(*ia));

	return (1);
}

/*
 * internal_authprov_reset_stats - reset authentication provider statistics
 */
static void
internal_authprov_reset_stats(struct authkey_be *be)
{
	struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;

	ibe->stats.keylookups	= 0;
	ibe->stats.keynotfound	= 0;
	ibe->stats.encryptions	= 0;
	ibe->stats.decryptions	= 0;
	ibe->stats.keyuncached	= 0;
}

/**** Keyfile authentication provider backend ****/

/*
 * keyfile_authprov_destructor - destructor for internal providers
 */
static void
keyfile_authprov_destructor(struct authkey_be *be)
{
	struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;

	if (ibe->keyfile != NULL) {
		free(ibe->keyfile);
		ibe->keyfile = NULL;
	}
	free(ibe);
}

/*
 * keyfile_auth_delkeys - delete all known keys, in preparation for rereading
 *			  the keys file (presumably)
 */
void
keyfile_authprov_delkeys(struct authkey_be *be)
{
	struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;
	struct savekey *sk;
	struct savekey **skp;
	int i;

	for (i = 0; i < HASHSIZE; i++) {
		skp = &(ibe->key_hash[i]);
		sk = ibe->key_hash[i];
		while (sk != 0) {
			/*
			 * Don't lose info as to which keys are trusted.
			 */
			if (sk->flags & KEY_TRUSTED) {
				skp = &(sk->next);
				memset(&sk->k, 0, sizeof(sk->k));
				sk->lifetime = 0;
				sk->keylen = 0;
				sk = sk->next;
			} else {
				*skp = sk->next;
				ibe->stats.numkeys--;
				sk->next = ibe->authfreekeys;
				ibe->authfreekeys = sk;
				ibe->stats.numfreekeys++;
				sk = *skp;
			}
		}
	}
}

/*
 * keyfile_authprov_refresh - reload keys from file
 */
static int
keyfile_authprov_refresh(struct authkey_be *be)
{
	struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;

	if (ibe->keyfile != NULL) {
		return keyfile_authprov_readkeys(be, ibe->keyfile);
	}

	return (1);
}

int
keyfile_authprov_setkey(
	struct authkey_be *be,
	keyid_t keyno,
	int type,
	const u_char *key,
	const int len,
	u_long trust
	)
{
	return internal_authprov_setkey(be, keyno, type, key, len, trust);
}

/*
 * keyfile_authprov_init - constructor for keyfile provider
 */
struct authkey_be *
keyfile_authprov_init(int version, int argc, const char *argv[])
{
	struct internal_authkey_be *keyfile_be;

	if (argc > 1) {
		return (NULL);
	}

	keyfile_be = (struct internal_authkey_be *)calloc(1, sizeof(*keyfile_be));
	if (keyfile_be == NULL) {
		return (NULL);
	}

	keyfile_be->be.destructor	= keyfile_authprov_destructor;
	keyfile_be->be.getmac		= internal_authprov_getmac;
	keyfile_be->be.verifymac	= internal_authprov_verifymac;
	keyfile_be->be.agekeys		= NULL; /* file keys do not age */
	keyfile_be->be.havekey		= internal_authprov_havekey;
	keyfile_be->be.trust		= internal_authprov_trust;
	keyfile_be->be.istrusted	= internal_authprov_istrusted;
	keyfile_be->be.setkey		= keyfile_authprov_setkey;
	keyfile_be->be.refresh		= keyfile_authprov_refresh;
	keyfile_be->be.stats		= internal_authprov_stats;
	keyfile_be->be.reset_stats	= internal_authprov_reset_stats;

	if (argc > 0) {
		keyfile_be->keyfile = strdup(argv[0]);
		if (keyfile_be->keyfile == NULL) {
			(*keyfile_be->be.destructor)((struct authkey_be *)keyfile_be);
			return (NULL);
		}

		/* Load keys from keyfile */
		if ((*keyfile_be->be.refresh)((struct authkey_be *)keyfile_be) == 0) {
			(*keyfile_be->be.destructor)((struct authkey_be *)keyfile_be);
			return (NULL);
		}
	}

	return (struct authkey_be *)keyfile_be;
}

/**** Autokey authentication provider backend ****/

#define AUTOKEY_BE_P(be)	((be) != NULL && (be)->destructor == autokey_authprov_destructor)

/*
 * autokey_authprov_destructor - destructor for internal providers
 */
static void
autokey_authprov_destructor(struct authkey_be *be)
{
	/* No keyfile */
	free(be);
}

/*
 * autokey_auth_agekeys - delete keys whose lifetimes have expired
 *
 * This is never called for keyfile keys because they do not expire
 */
static void
autokey_authprov_agekeys(struct authkey_be *be)
{
	struct internal_authkey_be *ibe = (struct internal_authkey_be *)be;
	struct savekey *sk;
	struct savekey *skp;
	int i;

	for (i = 0; i < HASHSIZE; i++) {
		sk = skp = ibe->key_hash[i];
		while (sk != 0) {
			skp = sk->next;
			if (sk->lifetime > 0 && current_time >
			    sk->lifetime) {
				internal_authprov_trust(be, sk->keyid, 0);
				ibe->stats.expired++;
			}
			sk = skp;
		}
	}
#ifdef DEBUG
	if (debug)
		printf("auth_agekeys: at %lu keys %u expired %u\n",
		    current_time, ibe->stats.numkeys, ibe->stats.expired);
#endif
}

/*
 * autokey_authprov_init - constructor for autokey provider
 */
struct authkey_be *
autokey_authprov_init(int version, int argc, const char *argv[])
{
	struct internal_authkey_be *autokey_be;

	if (argc != 0) {
		return (NULL);
	}

	autokey_be = (struct internal_authkey_be *)calloc(1, sizeof(*autokey_be));
	if (autokey_be == NULL) {
		return (NULL);
	}

	autokey_be->be.destructor	= autokey_authprov_destructor;
	autokey_be->be.getmac		= internal_authprov_getmac;
	autokey_be->be.verifymac	= internal_authprov_verifymac;
	autokey_be->be.agekeys		= autokey_authprov_agekeys;
	autokey_be->be.havekey		= internal_authprov_havekey;
	autokey_be->be.trust		= internal_authprov_trust;
	autokey_be->be.istrusted	= internal_authprov_istrusted;
	autokey_be->be.setkey		= internal_authprov_setkey;
	autokey_be->be.refresh		= NULL;
	autokey_be->be.stats		= internal_authprov_stats;
	autokey_be->be.reset_stats	= internal_authprov_reset_stats;

	return (struct authkey_be *)autokey_be;
}

/*
 * autokey_iskeyid - returns one if the key ID is associated with the
 *                   autokey authentication provider
 */
int
autokey_iskeyid(keyid_t keyid)
{
	struct authkey_be *be;

	be = authprov_select(keyid);

	return (AUTOKEY_BE_P(be));
}

/*
 * autokey_untrust - untrust an autokey with a single provider lookup
 */
void
autokey_untrust(keyid_t keyid)
{
	struct authkey_be *be;

	be = authprov_select(keyid);

	if (AUTOKEY_BE_P(be)) {
		be->trust(be, keyid, 0);
	}
}

